#ifndef __dltrans
#define __dltrans

#include <iostream>
#include <vector>
#include <cmath>

#include <blitz/array.h>
#include <cross.h>

template<class FT> class StraightSegment3D {
public:  
  StraightSegment3D(const blitz::Array<FT, 1>& c, bool _f = false) :
    fixed(_f), coordinates(c), dir(3), center(3), base(3, 3), burgers(3) { }
  StraightSegment3D(FT cx, FT cy, FT cz, bool _f = false) :
    fixed(_f), coordinates(3), dir(3), center(3), base(3, 3), burgers(3) {
    blitz::Array<FT, 1> coordinates;
    coordinates = cx, cy, cz;
  }
  const blitz::Array<FT, 1>& Fixed() const { return fixed; }
  const blitz::Array<FT, 1>& Coordinates() const { return coordinates; }
  blitz::Array<FT, 1>& Coordinates() { return coordinates; }
  const blitz::Array<FT, 1>& Center() const { return center; }
  blitz::Array<FT, 1>& Center() { return center; }
  FT Len() const { return len; }
  const blitz::Array<FT, 1>& Dir() const { return dir; }
  const blitz::Array<FT, 2>& Base() const { return base; }
  const blitz::Array<FT, 1>& Burgers() const { return burgers; }
  void ComputeTransformation3D(const StraightSegment3D<FT>& prev,
    const blitz::Array<FT, 1>& _burgers);
private:
  bool fixed;                      /* fixed point flag */
  blitz::Array<FT, 1> coordinates; /* segment point coordinates */
  FT len;                          /* segment length */
  blitz::Array<FT, 1> dir;         /* segment direction */
  blitz::Array<FT, 1> center;      /* segment center */
  blitz::Array<FT, 2> base;        /* segment base */
  blitz::Array<FT, 1> burgers;     /* segment burgers vector (rotated) */
};

/* This function constructs a transformation blitz::Array for segment of the
 * line. In fact, these transformation arrays consist of a rotation and a shift. The
 * rotation matrix is merely a base matrix constructed with z axis directed along the
 * dislocation segment. So it transforms vectors from the segment's base to the main
 * system base. The shift component is devised accordingly.
 * Note: This function provides both transformations. At first comes the transformation
 * main base -> segment base, then comes the backward transformation.
 * */
template<class FT> void StraightSegment3D<FT>::ComputeTransformation3D(
  const StraightSegment3D<FT>& prev, const blitz::Array<FT, 1>& _burgers) {
  FT size;
  blitz::Array<FT, 1> tmp(3);
  blitz::firstIndex I; blitz::secondIndex J;
  center = (Coordinates()(0) + prev.Coordinates()(0)) / 2, /* center of segment */
           (Coordinates()(1) + prev.Coordinates()(1)) / 2,
           (Coordinates()(2) + prev.Coordinates()(2)) / 2;
  /* base(2) is the direction vector, it directs new 'z' axis */
  dir = Coordinates() - prev.Coordinates();
  dir /= (len = sqrt(sum(sqr(dir))));
  base(blitz::Range::all(), 2) = dir;
  /* some normal vector (cross product of base(2) and Burgers vector */
  tmp = _burgers / sqrt(sum(sqr(_burgers)));
//  base(blitz::Range::all(), 1) = cross(tmp, base(blitz::Range::all(), 2));
  base(blitz::Range::all(), 1) = cross(base(blitz::Range::all(), 2), tmp);
#if 0  
  std::cout << "burgers: " << tmp << std::endl
            << "dir: "     << base(blitz::Range::all(), 2) << std::endl
            << "(b x d): " << base(blitz::Range::all(), 1) << std::endl;
#endif
  /* Burgers vector may be colinear with the direction vector */
  if ((size = sqrt(sum(sqr(base(blitz::Range::all(), 1))))) < 1e-7) { /* linear dependent */
    /* this construction is not ideal, we can unwind it :-) */
    /* tmp = base(1, 2), -base(2, 2), base(0, 2);
     * base(1, blitz::Range::all()) = cross(base(2, Range::all()), tmp); */
    base(0, 1) =  base(0, 2) * base(1, 2) + base(2, 2) * base(2, 2);
    base(1, 1) =  base(1, 2) * base(2, 2) - base(0, 2) * base(0, 2);
    base(2, 1) = -base(2, 2) * base(0, 2) - base(1, 2) * base(1, 2);
    base(blitz::Range::all(), 1) /= sqrt(sum(sqr(base(blitz::Range::all(), 1))));
  }
  /* another normal vector (cross product of base(1) and base(2)) */
//  base(blitz::Range::all(), 0) = cross(base(blitz::Range::all(), 1), base(blitz::Range::all(), 2));
  base(blitz::Range::all(), 0) = cross(base(blitz::Range::all(), 1), base(blitz::Range::all(), 2));
  base(blitz::Range::all(), 0) /= sqrt(sum(sqr(base(blitz::Range::all(), 0))));
  base(blitz::Range::all(), 1) /= sqrt(sum(sqr(base(blitz::Range::all(), 1))));

  /* Now we have the transformation, we can transform the burgers vector. */
//  base.transposeSelf(blitz::secondDim, blitz::firstDim);
  burgers = sum(base(J, I) * _burgers(J), J);
}

/* This function iterates over dislocation lines and construct transformation matrices. */
template<typename FT> void ComputeDLTransformation3D(
  std::vector<StraightSegment3D<FT> >& straightsegments,
  const blitz::Array<FT, 1>& burgers) {
  typename std::vector<StraightSegment3D<FT> >::iterator cc, cn;
  cn = straightsegments.begin();
  cc = cn++;
  for (; cn != straightsegments.end(); ++cn) {
    cn->ComputeTransformation3D(*cc, burgers);
    cc = cn;
  }
}

#endif /* __dltrans */
